/*
   This program tries to verify the picture that we drew in class
   of how condition variables work.

   This program shows how threads in the lock queue can enter that
   queue from different places and then return to those different
   places when those threads leave the lock queue.
*/

import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Condition;

public class Test
{
   public static final ReentrantLock lock = new ReentrantLock();
   public static final Condition condition1 = lock.newCondition();
   public static final Condition condition2 = lock.newCondition();

   public static void main(String[] args)
   {
      try
      {
         Runnable r1 = new Runnable1();
         Runnable r2 = new Runnable2();
         Runnable r3 = new Runnable3();
         Runnable r4 = new Runnable4();
         Runnable r5 = new Runnable5();

         Thread t1 = new Thread(r1);
         Thread t2 = new Thread(r2);
         Thread t3 = new Thread(r3);
         Thread t4 = new Thread(r4);
         Thread t5 = new Thread(r5);

         t1.start();  Thread.sleep(1);
         t2.start();  Thread.sleep(1);
         t3.start();  Thread.sleep(1);
         t4.start();  Thread.sleep(1);
         t5.start();  Thread.sleep(1);

         t3.interrupt();
      }
      catch (InterruptedException e)
      {
      }
   }
}

class Runnable1 implements Runnable
{
   @Override
   public void run()
   {
      try
      {
         System.out.println("Thread 1 acquires the lock.");
         Test.lock.lock();
         System.out.println("Thread 1 waiting on condition 1 (releases the lock).");
         Test.condition1.await();
         System.out.println("Thread 1 reacquires the lock (done waiting on condition 1).");
      }
      catch (InterruptedException e)
      {
         System.out.println("Thread 1 interrupted (this shouldn't happen).");
      }
      finally
      {
         System.out.println("Thread 1 releases the lock.");
         Test.lock.unlock();
      }
   }
}

class Runnable2 implements Runnable
{
   @Override
   public void run()
   {
      try
      {
         System.out.println("Thread 2 acquires the lock.");
         Test.lock.lock();
         System.out.println("Thread 2 waiting on condition 2 (releases the lock).");
         Test.condition2.await();
         System.out.println("Thread 2 reacquires the lock (done waiting on condition 2).");
      }
      catch (InterruptedException e)
      {
         System.out.println("Thread 2 interrupted (this shouldn't happen).");
      }
      finally
      {
         System.out.println("Thread 2 releases the lock.");
         Test.lock.unlock();
      }
   }
}

class Runnable3 implements Runnable
{
   @Override
   public void run()
   {
      try
      {
         System.out.println("Thread 3 acquires the lock.");
         Test.lock.lock();
         while (! Thread.interrupted())
         {
            Thread.sleep(1000);
         }
      }
      catch (InterruptedException e)
      {
         System.out.println("Thread 3 interrupted.");
      }
      finally
      {
         System.out.println("Thread 3 signals Thread 2.");
         Test.condition2.signalAll();
         System.out.println("Thread 3 signals Thread 1.");
         Test.condition1.signalAll();
         try
         {
            System.out.println("Threads 1, 2, 4, 5 are all in the lock queue.");
            System.out.println("Thread 3 sleeping for 3 seconds.");
            Thread.sleep(3_000);
         }
         catch (InterruptedException e){}
         System.out.println("Thread 3 releases the lock.");
         Test.lock.unlock();
      }
   }
}

class Runnable4 implements Runnable
{
   @Override
   public void run()
   {
      try
      {
         System.out.println("Thread 4 requests the lock.");
         Test.lock.lock();
      }
      finally
      {
         System.out.println("Thread 4 releases the lock.");
         Test.lock.unlock();
      }
   }
}

class Runnable5 implements Runnable
{
   @Override
   public void run()
   {
      try
      {
         System.out.println("Thread 5 requests the lock.");
         Test.lock.lock();
      }
      finally
      {
         System.out.println("Thread 5 releases the lock.");
         Test.lock.unlock();
      }
   }
}
